-- LJ Sonic's hub


local eventdefs


freeslot("MT_LJHUB_NPC", "S_LJHUB_NPC")

mobjinfo[MT_LJHUB_NPC] = {
	doomednum = 3470,
	spawnstate = S_LJHUB_NPC,
	spawnhealth = 1,
	radius = 16 * FRACUNIT,
	height = 56 * FRACUNIT,
	flags = 0
}

states[S_LJHUB_NPC] = {SPR_PLAY, SPR2_STND|FF_ANIMATE, -1, nil, 0, 7, S_NULL}


local function getEventNpc(mo, maxdist)
	local MT_LJHUB_NPC = MT_LJHUB_NPC
	local bestnpc, bestdist = nil, maxdist

	for npc in mobjs.iterate() do
		if npc.type == MT_LJHUB_NPC then
			local dist = R_PointToDist2(mo.x, mo.y, npc.x, npc.y)
			if dist < bestdist then
				bestnpc, bestdist = npc, dist
			end
		end
	end

	return bestnpc
end

local function setQuest(p, quest)
	p.ljhub_quest = ("Quest:\n" .. quest)
end

local function stopEvent(p)
	local event = p.ljhub_event
	local eventdef = eventdefs[event.id]

	if not eventdef.noprogress and eventdef.story ~= nil
	and not eventdef.alone then
		p.ljhub_story = eventdef.nextstory or $ + 1
	end

	if not eventdef.alone and eventdef.quest then
		setQuest(p, eventdef.quest)
	end

	if eventdef.once then
		p.ljhub_uniqueeventsdone[event.id] = true
	end

	event.closed = true
end

local function runEvent(p)
	local event = p.ljhub_event
	local eventdef = eventdefs[event.id]

	while event.line ~= #eventdef do
		event.line = $ + 1
		local line = eventdef[event.line]

		if type(line) == "string" then
			event.pos = line:find("\n") * FRACUNIT
			return
		elseif type(line) == "function" then
			line(p)
		end
	end

	stopEvent(p)
end

local function startEvent(p, eventid)
	local eventdef = eventdefs[eventid]

	local npc
	if not (eventdef.alone or eventdef.sign) then
		npc = getEventNpc(
			p.mo,
			eventdef.nopresencecheck and INT32_MAX or 1024 * FRACUNIT
		)
	end

	p.ljhub_event = {
		id = eventid,
		line = 0,
		stoptime = (eventdef.sign and TICRATE / 2 or 3 * TICRATE),
		npc = npc
	}

	if eventdef.alone then
		if eventdef.quest then
			setQuest(p, eventdef.quest)
		end

		if not eventdef.noprogress and eventdef.story ~= nil then
			p.ljhub_story = eventdef.nextstory or $ + 1
		end
	end

	runEvent(p)
end


addHook("PreThinkFrame", function()
	if not mapheaderinfo[gamemap].ljhub then return end

	for p in players.iterate do
		-- Foolproof.
		if p.rings < 1 then
			P_GivePlayerRings(p, 1)
		end

		if p.ljhub_ink then
			if P_RandomChance(FRACUNIT / (TICRATE / 2)) then
				local goop = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_GFZFLOWER1)
				goop.sprite, goop.frame = SPR_GOOP, C
				goop.colorized = true
				goop.color = SKINCOLOR_BLACK
				goop.fuse = 30 * TICRATE
				goop.tics = goop.fuse + 1
			end

			p.ljhub_ink = $ - 1
			if p.ljhub_ink == 0 then
				p.ljhub_ink = nil
			end
		end

		local event = p.ljhub_event
		if not event then continue end
		local eventdef = eventdefs[event.id]

		if event.closed then
			if p.cmd.buttons & BT_JUMP then
				p.cmd.buttons = $ & ~BT_JUMP
			else
				p.ljhub_event = nil
			end
			continue
		end

		if not (eventdef.nopresencecheck or eventdef.alone) then
			local npc = event.npc
			if npc then
				local dist = R_PointToDist2(p.mo.x, p.mo.y, npc.x, npc.y)
				if dist > 1024 * FRACUNIT then
					startEvent(p, 2001)
					p.ljhub_event.npc = npc
					continue
				end
			else
				event.stoptime = $ - 1
				if event.stoptime == 0 then
					if eventdef.sign then
						p.ljhub_event = nil
					else
						startEvent(p, 2002)
					end
					continue
				end
			end
		end

		event.pos = $ + FRACUNIT

		local skip = p.cmd.buttons & BT_JUMP
			and event.prevbuttons ~= nil
			and not (event.prevbuttons & BT_JUMP)

		event.prevbuttons = p.cmd.buttons
		p.cmd.buttons = $ & ~BT_JUMP

		if skip then
			local line = eventdef[event.line]
			if event.pos < #line * FRACUNIT then
				event.pos = #line * FRACUNIT
			else
				runEvent(p)
			end
		end
	end
end)

addHook("MobjSpawn", function(mo)
	local skins = {
		"sonic",
		"knuckles",
		"amy",
		"fang",
		"metalsonic",
	}

	mo.skin = skins[P_RandomRange(1, #skins)]
	mo.color = P_RandomRange(1, #skincolors - 1)
end, MT_LJHUB_NPC)

addHook("LinedefExecute", function(l, mo)
	local p = mo.player
	if not p then return end

	local currentevent = p.ljhub_event
	local currenteventdef
	if currentevent
		currenteventdef = eventdefs[currentevent.id]
	end

	local eventid = l.frontside.textureoffset / FRACUNIT
	local eventdef = eventdefs[eventid]
	if not eventdef then return end

	p.ljhub_story = $ or 1
	--p.ljhub_story = $ or 21 -- !!! dbg
	p.ljhub_uniqueeventsdone = $ or {}

	if eventdef.story ~= nil
	and not (p.ljhub_story >= eventdef.story[1] and p.ljhub_story <= eventdef.story[2]) then
		return
	elseif eventdef.once and p.ljhub_uniqueeventsdone[eventid] then
		return
	elseif p.ljhub_event and p.ljhub_event.id == eventid then
		p.ljhub_event.stoptime = (eventdef.sign and TICRATE / 2 or 3 * TICRATE)
		return
	elseif currentevent and not currenteventdef.sign then
		return
	else
		startEvent(p, eventid)
	end
end, "LJHUBEVENT")

addHook("LinedefExecute", function(_, mo)
	local p = mo.player
	if not p then return end

	p.ljhub_ink = ($ or 0) + 4
	p.ljhub_ink = min(max($, 5 * TICRATE), 60 * TICRATE)
end, "LJHUBINK")

addHook("MapChange", function()
	if not mapheaderinfo[gamemap].ljhub then return end

	for p in players.iterate do
		p.ljhub_event = nil
		p.ljhub_story = nil
		p.ljhub_uniqueeventsdone = nil
		p.ljhub_quest = nil
	end
end)

local border = 2
local gap = 2

local w = 320 - 2 * gap
local h = 52
local left = gap
local top = 200 - gap - h
local right = left + w - 1
local bottom = top + h - 1

local function drawBox(v)
	v.drawFill(left, top, w, h, 0)
	v.drawFill(left + border, top + border, w - 2 * border, h - 2 * border, 31)
end

local function drawFace(v, npc)
	local sprite = v.getSprite2Patch(npc.skin, SPR2_SIGN)
	local colormap = v.getColormap(nil, npc.color)

	local x = right - border - sprite.width + sprite.leftoffset
	local y = bottom - border

	v.draw(x, y, sprite, 0, colormap)
end

local function drawQuest(v, quest)
	quest = $:gsub("\t", "")

	local y = 2
	for line in quest:gmatch("[^\n]+") do
		v.drawString(318, y, line, V_SNAPTORIGHT | V_SNAPTOTOP | V_ALLOWLOWERCASE, "right")
		y = $ + 10
	end
end

hud.add(function(v, p)
	if not mapheaderinfo[gamemap].ljhub then return end

	if p.ljhub_quest then
		drawQuest(v, p.ljhub_quest)
	end

	local event = p.ljhub_event
	if not event or event.closed return end

	local s = eventdefs[event.id][event.line]
	if type(s) ~= "string" then return end

	drawBox(v)

	if s:sub(1, 4) == "you\n" then
		drawFace(v, p.mo)
	elseif s:sub(1, 4) == "npc\n" and p.mo then
		if event.npc then
			drawFace(v, event.npc)
		end
	end

	-- Text
	s = s:sub(s:find("\n") + 1, event.pos / FRACUNIT)
	v.drawString(left + border + 3, top + border + 2, s, V_ALLOWLOWERCASE)
end)


eventdefs = {
[1] = {
	quest = [[
		3rd floor
		Left corridor
		1st door
	]],

	[[
		npc
		Hello!
		Do you need help with something?
	]],
	[[
		you
		Yes, where are the levels?
	]],
	[[
		npc
		The levels?
		They are just behind this gate.
	]],
	[[
		npc
		However you are required to
		fill a special form in order
		to be allowed to pass through.
	]],
	[[
		npc
		Please come back with a form,
		and I will let you
		go to the levels.
	]],
	[[
		you
		A form?
		Where do I find that?
	]],
	[[
		npc
		You should find one on the
		third floor, corridor on the left,
		first door. Just ask them for
		a form, they will understand.
	]],
	[[
		you
		Understood.
		Thank you!
	]],
	[[
		npc
		Have a nice day!
	]],
},

[2] = {
	alone = true,
	quest = "???",

	[[
		you
		Corridor on the left...
		Hm...
	]],
	[[
		you
		There are no doors here.
	]],
},

[3] = {
	alone = true,

	quest = [[
		1st floor
		White gate
	]],

	[[
		you
		I guess I should come back
		and ask them again.
	]],
},

[4] = {
	quest = "5th floor",

	[[
		you
		There must be an error.
		There were no doors
		in that corridor.
	]],
	[[
		npc
		Really? That's strange.
		I'm not sure where you are
		supposed to go, then.
	]],
	[[
		npc
		You should ask the secretary
		where you can find the forms.
	]],
	[[
		npc
		The secretary is on the
		third floor, at the last door
		of the left corridor.
	]],
	[[
		you
		...Wait, this is where I come from.
		I told you there is nothing there.
	]],
	[[
		npc
		Ah, indeed.
		I forgot.
	]],
	[[
		npc
		Then hm...
	]],
	[[
		npc
		I know, you should consult the map.
		There is a map of the building
		somewhere on the fifth floor.
	]],
	[[
		you
		Got it, thank you again!
	]],
	[[
		npc
		You're welcome!
	]],
},

[5] = {
	quest = "...",

	[[
		you
		Ah, the map of the building.
		Finally found it.
	]],
	[[
		you
		Let's see...
	]],
	[[
		you
		So, the top is here... no, here.
		And the bottom...
		wait no, it's the top?
	]],
	[[
		you
		Oh okay, actually it's the
		last floor. The first floor is
		on the leftmost area.
	]],
	[[
		you
		And then it says the forms are
		at the second... no, third,
		no actually, it's underground.
	]],
	[[
		you
		Ah no, it's not underground,
		it's the roof... uh????
		Ah nevermind, it's the street.
	]],
	[[
		you
		Then where are the forms hm.
		It says this area here is the...
		uh... seventeenth floor??
	]],
	[[
		you
		Hm... Ah okay, it's another
		building. But this area
		must be the... wait no.
	]],
	[[
		you
		Okay, let's start from scratch.
		So, the room where you can find
		the forms is here.
	]],
	[[
		you
		Ah, it says "only on rainy days".
		Weird.
		But then I guess it's...
		uh... wait...
	]],
	[[
		you
		...why is the sixth floor below
		the fourth floor????
	]],
	[[
		you
		Oh I see, it's not a floor,
		it's a ceiling. And the roof
		is underground... what...??
	]],
	[[
		you
		Hm...
	]],
	[[
		you
		Hmmmmmm......
	]],
	[[
		you
		Hm.
	]],
	[[
		you
		HM.
	]],
	[[
		you
		HHMMHHMMHaa?E??as3SDF!!s
		d,66F4a.fss!!FQBSS,,.?ASQ
		ZARGJGJ?????????????
	]],
	[[
		you
		THIS MAP.
	]],
	[[
		you
		WHAT IS THIS MAP.
	]],
},

[6] = {
	alone = true,
	quest = [[
		1st floor
		White gate
	]],

	[[
		you
		...
	]],
	[[
		you
		Well.
	]],
	[[
		you
		I guess I will come back
		one more time and ask
		for help again...
		Again...
	]],
},

[7] = {
	quest = [[
		2nd floor
		Corridor F
		Last door
	]],

	[[
		npc
		Oh, you're back.
		Did you get a form?
	]],
	[[
		you
		Well...
	]],
	[[
		you
		I managed to find the map,
		but I am unable to make
		any sense of it...
	]],
	[[
		npc
		Ah?
		Well, I guess it can be a little
		confusing to some people.
	]],
	[[
		you
		...Yes...
	]],
	[[
		you
		By the way, why are the stairs
		so weirdly shaped...?
	]],
	[[
		npc
		Ah, that?
	]],
	[[
		npc
		When we asked her, the architect
		mentioned something about
		"FoFs being a pain in the ass",
		or something like that.
	]],
	[[
		npc
		Nobody really understood what
		she meant, though.
	]],
	[[
		npc
		Anyway, can you quickly go look
		at the map again, and tell me
		what labels are written inside
		area G-27 and ar-
	]],
	[[
		you
		NO.
	]],
	[[
		npc
		...C-Calm down, please.
		I was just-
	]],
	[[
		you
		Sorry, I mean, I'm bad with maps,
		so I'd rather ask help from someone?
	]],
	[[
		npc
		I understand.
		Then, go to the third floor,
		corridor on the left, and-
	]],
	[[
		you
		No.
	]],
	[[
		npc
		Right, right.
		Then, second floor, corridor F,
		the door at the end of the corridor.
	]],
	[[
		you
		I will go there... I guess...
	]],
	[[
		npc
		You're welcome!
	]],
},

[8] = {
	[[
		you
		Excuse me, I am trying to obtain
		a form to access the levels.
	]],
	[[
		npc
	]],
	[[
		npc
	]],
	[[
		you
		...Hello?
	]],
	[[
		npc
	]],
	[[
		you
		HELLO?
	]],
	[[
		npc
		?!
	]],
	[[
		npc
		What are you doing here??
	]],
	[[
		you
		I... The form...
		I need a form.
	]],
	[[
		npc
		I am busy.
	]],
	[[
		you
		...
	]],
	[[
		npc
		And *PLEASE* don't forget
		to close the door.
	]],
	[[
		npc
		(God, people these days...)
	]],
},

[9] = {
	nopresencecheck = true,
	quest = "5th floor",

	[[
		npc
		Wait!
	]],
	[[
		npc
		I know how you can find forms.
	]],
	[[
		npc
		There is a map of the building
		on the fifth floor.
	]],
},

[10] = {
	alone = true,

	function(p)
		P_LinedefExecute(42, p.mo)
		p.ljhub_mazestart = leveltime
	end,
},

[11] = {
	alone = true,

	[[
		you
		...
	]],
},

[12] = {
	story = 12,
	nextstory = 14,
	quest = "Office 2",

	[[
		npc
		Hello!
		May I help you?
	]],
	[[
		you
		Do you know where I can obtain
		the form required to be allowed
		through the white gate?
	]],
	[[
		npc
		A form?
		Please ask my colleague
		at office 2.
	]],
	[[
		you
		Oh.
	]],
},

[13] = {
	story = 12,
	nextstory = 14
},

[14] = {
	quest = "Office 1",

	[[
		npc
		Hello!
		Can I help you with anything?
	]],
	[[
		you
		So, do you know where I can get
		a form for the levels?
	]],
	[[
		npc
		Oh, you need a form?
		You should ask at office 1,
		my colleague should be able
		to print one for you.
	]],
},

[15] = {
	quest = "Office 3",

	[[
		you
		Your colleague told me
		you can make forms.
	]],
	[[
		npc
		That must be an error, for
		forms you should be asking
		at office 3.
	]],
},

[16] = {
	quest = "Office 1",
},

[17] = {
	quest = "??????",

	[[
		you
		...Office 3 is closed.
	]],
	[[
		npc
		Closed?
	]],
	[[
		you
		Yes.
	]],
	[[
		npc
		Oh, you mean office 38.
	]],
	[[
		you
		38??
	]],
	[[
		npc
		The office next to office 2
		is office 38.
	]],
	[[
		npc
		Office 3 is on floor 12,
		corridor W, door 17, building E.
	]],
	[[
		you
		??????
	]],
},

[18] = {
	alone = true,
	nextstory = 21,

	function(p)
		P_LinedefExecute(47, p.mo)
	end,
},

[19] = {
	story = 21,
	once = true,
	noprogress = true,

	[[
		you
		Excuse me...
	]],
	[[
		npc
		I'm very busy,
		please ask my colleague.
	]],
},

[21] = {
	quest = [[
		3rd floor
		Opposite desk
	]],

	[[
		npc
		Hello, are you looking for someone?
	]],
	[[
		you
		Hm...
		Where am I?
	]],
	[[
		npc
		Where? What do you mean?
		Well, we're in my office.
		On the third floor.
	]],
	[[
		you
		I... see.
	]],
	[[
		npc
		Are you lost?
		Maybe I can help.
	]],
	[[
		you
		I... I'm not even sure anymore.
	]],
	[[
		npc
		?
	]],
	[[
		you
		I think I'm looking
		for... building E??
		What building are we in anyway.
	]],
	[[
		npc
		We're in building A.
		I'm not sure where building E
		is, but you should find a ma-
	]],
	[[
		you
		Where do I find forms.
	]],
	[[
		npc
		Forms?
		What kind of forms?
	]],
	[[
		you
		I want to leave this building
		and play the levelpack.
		But I need a form to pass
		through the exit gate.
	]],
	[[
		npc
		Oh.
		I can give you one.
	]],
	[[
		you
		Wait, really?
	]],
	[[
		npc
		Sure thing.
		I can even fill it for you,
		if you want.
		Let me do that...
	]],
	[[
		you
		That's great.
		Thank you.
	]],
	[[
		npc
		Wait a moment please...
	]],
	[[
		npc
		...
	]],
	[[
		npc
		...
		...
	]],
	[[
		npc
		...
		...
		...
	]],
	[[
		npc
		...
		...
		...
		...
	]],
	[[
		npc
		Do you have a yellow ticket?
	]],
	[[
		you
		A what?
	]],
	[[
		npc
		I can't give you a form if
		you don't have a yellow ticket.
	]],
	[[
		you
		How do I get that.
	]],
	[[
		npc
		I don't remember...
		You should ask at
		the opposite desk.
	]],
},

[22] = {
	story = 22,
	nextstory = 24,

	quest = [[
		1st floor
		White gate
	]],

	[[
		you
		Excuse me.
	]],
	[[
		npc
		What?
	]],
	[[
		you
		I need a yellow ticket.
		How do I get one?
	]],
	[[
		npc
		Ask the receptionist.
	]],
	[[
		npc
		Now if you'll excuse me.
	]],
},

[23] = {
	story = 22,
	nextstory = 24,
},

[24] = {
	quest = [[
		1st floor
		Left corridor
	]],

	[[
		npc
		You again?
		What were you doing?
	]],
	[[
		npc
		Have you got a form now?
	]],
	[[
		you
		No.
	]],
	[[
		you
		I need a yellow ticket now.
	]],
	[[
		npc
		A yellow ticket?
		Let me check...
	]],
	[[
		npc
		Apparently they distribute
		tickets at office D-1-3-7bis.
	]],
	[[
		you
		Where is that.
	]],
	[[
		npc
		Left corridor, right after
		exiting the entrance hall.
	]],
	[[
		you
		On... this floor...?
	]],
	[[
		npc
		First floor, yes.
	]],
	[[
		you
		Ah.
	]],
},

[25] = {
	quest = [[
		1st floor
		White gate
	]],

	[[
		npc
		Hello!
		May I be of help?
	]],
	[[
		you
		Is this where I can
		get a yellow ticket?
	]],
	[[
		npc
		Yes.
		Please wait a moment...
	]],
	[[
		npc
		...
	]],
	[[
		npc
		There you go!
		Here is your green ticket.
	]],
	[[
		you
		Ah, thank you.
	]],
	[[
		npc
		You're welcome!
	]],
},

[26] = {
	alone = true,

	[[
		you
		...
	]],
	[[
		you
		Green?
	]],
},

[27] = {
	quest = [[
		1st floor
		Right corridor
	]],

	[[
		you
		They gave me a green ticket.
		Is that enough?
	]],
	[[
		npc
		No, no.
		You need a yellow ticket.
	]],
	[[
		you
		But...
	]],
	[[
		npc
		Try the other corridor,
		they may have moved there.
	]],
	[[
		you
		What do I do with that
		green ticket though?
	]],
	[[
		npc
		Oh.
		Give it to me, I'll bring
		them back for you.
	]],
},

[28] = {
	quest = [[
		1st floor
		White gate
	]],

	[[
		you
		Do you have yellow tickets.
	]],
	[[
		npc
		Yes.
		Do you have a green ticket?
	]],
	[[
		you
		No, I need a yellow ticket.
		Not green.
	]],
	[[
		npc
		I know, but I can only deliver
		yellow tickets if you have
		a green ticket.
	]],
	[[
		you
		...
	]],
	[[
		npc
		Are you okay?
	]],
	[[
		you
	]],
	[[
		npc
		...?
	]],
},

[29] = {
	quest = [[
		1st floor
		Left corridor
	]],

	[[
		npc
		Yes?
	]],
	[[
		you
		Please give me back the
		green ticket.
	]],
	[[
		npc
		I sent it back already,
		please ask for another one.
	]],
},

[30] = {
	quest = [[
		1st floor
		Right corridor
	]],

	[[
		npc
		Do you need something else?
	]],
	[[
		you
		Yes.
		Another ticket please.
	]],
	[[
		npc
		Oh.
	]],
	[[
		npc
		There you go!
	]],
},

[31] = {
	quest = [[
		3rd floor
		Somewhere
	]],

	[[
		npc
		Oh, you got a green ticket?
		Perfect.
	]],
	[[
		you
		Can I have the yellow one now?
	]],
	[[
		npc
		Of course.
		Please give me a moment...
	]],
	[[
		npc
		Let's see...
		Ticket... ticket...
	]],
	[[
		npc
		...
	]],
	[[
		npc
		...I'm out of tickets.
	]],
	[[
		you
		.
	]],
	[[
		npc
		Oh, nevermind.
		I was looking at the wrong place.
	]],
	[[
		npc
		There, your ticket.
	]],
	[[
		you
		...Thank you...
	]],
	[[
		npc
		You're welcome!
	]],
},

[32] = {
	alone = true,

	[[
		you
		The room was behind
		one of these doors...
	]],
},

[33] = {
	quest = [[
		1st floor
		White gate
	]],

	[[
		npc
		Did you get the ticket?
	]],
	[[
		you
		Yes... can I get the form?
	]],
	[[
		npc
		Please sign here.
	]],
	[[
		npc
		And also here.
	]],
	[[
		npc
		And here.
	]],
	[[
		npc
		And here.
	]],
	[[
		npc
		And here.
	]],
	[[
		npc
		And here.
	]],
	[[
		npc
		And here.
	]],
	[[
		npc
		And here.
	]],
	[[
		npc
		Thank you.
	]],
	[[
		you
		...That's... that's all?
	]],
	[[
		npc
		Well, yes.
	]],
	[[
		npc
		Normally you would also need
		a red ticket and a blue ticket,
		but we have to rush it because
		of the submission deadline.
	]],
	[[
		you
		??
	]],
	[[
		you
		Anyway.
		Thank you for the form!
	]],
	[[
		npc
		You're welcome.
		Have a good evening!
	]],
},

[34] = {
	quest = "Closed.",

	[[
		you
		I got the form!
		Can I pass through?
	]],
	[[
		npc
		I'm sorry, we are closing.
	]],
	[[
		you
		What?
	]],
	[[
		you
		Wait, I have the form,
		can't I just pass before
		you close?
	]],
	[[
		npc
		Please come back tomorrow.
	]],
	function(p)
		P_LinedefExecute(66)
	end,
},

[35] = {
	alone = true,

	function(p)
		local lighttags = {
			[64] = true,
			[72] = true,
			[73] = true,
		}

		local typestoremove = {
			[MT_FLAME] = true,
			[MT_PARTICLE] = true,
			[MT_FIRETORCH] = true,
			[MT_LJHUB_NPC] = true,
		}

		for s in sectors.iterate do
			if not lighttags[s.tag] then
				s.lightlevel = min($, 128)
			end
		end

		local lightstoremove = {}
		for mo in mobjs.iterate() do
			if typestoremove[mo.type] then
				table.insert(lightstoremove, mo)
			else
				mo.frame = $ & ~FF_FULLBRIGHT
			end
		end

		for _, light in ipairs(lightstoremove) do
			P_RemoveMobj(light)
		end

		-- Night sky
		P_SetupLevelSky(8)

		-- Bright gate turns black
		local l = lines[P_FindSpecialLineFromTag(0, 65)]
		l.frontside.midtexture = R_CheckTextureNumForName("~PIT")

		S_StartSound(nil, sfx_s3k98)

		-- Open elevator
		P_LinedefExecute(71)
	end,
},

[1001] = {
	sign = true,
	[[
		sign
		Corridors A to E ->
	]],
},

[1002] = {
	sign = true,
	[[
		sign
		<- Corridors G to K
	]],
},

[1003] = {
	sign = true,
	[[
		sign
		Corridor A
	]],
},

[1004] = {
	sign = true,
	[[
		sign
		Corridor B
	]],
},

[1005] = {
	sign = true,
	[[
		sign
		Corridor C
	]],
},

[1006] = {
	sign = true,
	[[
		sign
		Corridor D
	]],
},

[1007] = {
	sign = true,
	[[
		sign
		Corridor E
	]],
},

[1008] = {
	sign = true,
	[[
		sign
		Corridor F
	]],
},

[1009] = {
	sign = true,
	[[
		sign
		Corridor G
	]],
},

[1010] = {
	sign = true,
	[[
		sign
		Corridor H
	]],
},

[1011] = {
	sign = true,
	[[
		sign
		Corridor I
	]],
},

[1012] = {
	sign = true,
	[[
		sign
		Corridor J
	]],
},

[1013] = {
	sign = true,
	[[
		sign
		Corridor K
	]],
},

[1014] = {
	sign = true,
	[[
		sign
		- OPENING HOURS -
	]],
	[[
		sign
		(The rest of the paper is blank.)
	]],
},

[1015] = {
	sign = true,
	[[
		sign
		IT IS STRICTLY PROHIBITED TO
		STEP ON THE FLOWER.
	]],
},

[1016] = {
	sign = true,
	[[
		sign
		This paper contains text.
	]],
	[[
		sign
		The text is made of
		various characters.
	]],
	[[
		sign
		The characters are letters
		for the most part, but there are
		also some digits, punctuation
		marks and, of course, spaces.
	]],
	[[
		sign
		Most of the the letters are
		lowercase, although a few of
		them are uppercase, mostly
		at the start of each sentence.
	]],
	[[
		sign
		It appears you have reached
		the end of the paper.
		That was a nice read.
	]],
},

[1017] = {
	sign = true,
	[[
		sign
		- EVENTS OF THE MONTH -
	]],
	[[
		sign
		 1st: First-day-of-the-month party
		 2nd: Second-day-of-the-month party
		 7th: Day of the Port Forward
		13th: OLDC release
	]],
	[[
		sign
		17th: Half-month party
		21th: Thok debate (IMPORTANT)
		23th: Buy food for chao
		26th: Official Hangout Contest
	]],
	[[
		sign
		29th: 2.2.9 prerelease (cancelled)
		31th: Last-day-of-the-month party
		32th: Half-Life 38 release
	]],
	[[
		sign
		Can someone reach the paper above?
		It's too high for me.
	]],
},

[1018] = {
	sign = true,
	[[
		sign
		Did you know?
		You don't?
		Well, me neither.
	]],
},

[1019] = {
	sign = true,
	[[
		sign
		99 little bugs in the code.
		99 little bugs.
		Use regular expressions.
		100 little bugs in the code...
	]],
},

[1020] = {
	sign = true,
	[[
		sign
		READING THIS TEXT IS
		STRICTLY PROHIBITED.
	]],
},

[1021] = {
	sign = true,
	[[
		sign
		- CUSTOMER CHEATSHEET -
	]],
	[[
		sign
		1. Be courteous.
		2. Listen carefully.
		3. Do not get irritated/impatient.
		4. Act professional.
	]],
	[[
		sign
		5. If the customer needs information
		   regarding form obtention,
		   redirect them to office 1.
	]],
	[[
		sign
		6. If they have a question about
		   a rejected submission, redirect
		   them to the judges departement.
	]],
	[[
		sign
		7. If they need support with
		   port forwarding, redirect them
		   to the technical department.
	]],
	[[
		sign
		8. If they want to file a complaint
		   against thok removal, redirect
		   them to office 23. (3rd floor, left
		   corridor, last door on the right)
	]],
},

[1022] = {
	alone = true,

	function(p)
		if p.ljhub_mazestart ~= nil
		and leveltime > p.ljhub_mazestart + 15 * 60 * TICRATE then
			p.ljhub_mazestart = nil
		else
			stopEvent(p)
		end
	end,
	[[
		you
		...Well.
	]],
	[[
		you
		I'm completely lost.
	]],
	[[
		you
		I should probably go downstairs
		and find some help...
		before my sanity gets
		completely drained.
	]],
	function(p)
		setQuest(p, "Go downstairs")
	end,
},

[1023] = {
	story = 1,
	once = true,
	noprogress = true,

	quest = [[
		1st floor
		White gate
	]],

	[[
		npc
		Hello, do you need
		help with something?
	]],
	[[
		you
		Maybe.
		I am looking for the levels.
	]],
	[[
		npc
		Oh.
	]],
	[[
		npc
		The levels are behind the gate
		at the end of the main corridor.
		Talk to the receptionist next to
		the gate, they can open it.
	]],
	[[
		you
		Thank you!
	]],
},

[1025] = {
	story = 27,
	noprogress = true,
	once = true,

	[[
		npc
		Yes?
	]],
	[[
		you
		I wanted a yellow ticket.
	]],
	[[
		npc
		I only have green tickets.
		I don't think that matters.
	]],
	[[
		you
		Ah...
	]],
},

[1026] = {
	story = {1, 19},
	noprogress = true,
	once = true,

	[[
		you
		Excuse me...
	]],
	[[
		npc
		I am working.
		Please ask my colleague.
	]],
	[[
		you
		Oh, sorry.
	]],
},

[1027] = {
	story = {1, 19},
	noprogress = true,
	once = true,

	[[
		you
		Excuse me...
	]],
	[[
		npc
		Who are you?
		If you are looking
		for something, please
		ask somewhere else.
	]],
	[[
		you
		Ah, sorry.
	]],
},

[1028] = {
	story = 36,
	noprogress = true,
	once = true,
	alone = true,

	[[
		you
		Seems like the door is closed.
		They didn't even wait for me...
	]],
},

[1029] = {
	sign = true,
	[[
		sign
		   - IMPORTANT -
		- SAFETY NOTICE -
	]],
	[[
		sign
		1) Max. 32 players
		2) No smoking
		3) No campfires
		4) No detons
	]],
	[[
		sign
		5) No TNT
		6) No bomb rings
		7) No nuclear weapons
		8) No swearing!!!!!
	]],
	[[
		sign
		9) By the way a door has
		     been opened somewhere lol
	]],
	[[
		sign
		10) By the way a door has
		       been opened somewhere lol
	]],
	[[
		sign
		11) By the way a door has
		       been opened somewhere lol
	]],
},

[1030] = {
	sign = true,
	[[
		sign
		<-- Levels
	]],
},

[1031] = {
	sign = true,
	[[
		sign
		Bonus levels -->
	]],
},

[2001] = {
	nopresencecheck = true,

	[[
		npc
		Hey?
		Where are you going??
	]],
	[[
		npc
		Come here please...
	]],
},

[2002] = {
	alone = true,

	[[
		you
		Where I am going?
	]],
	[[
		you
		I should come back,
		I wasn't done yet...
	]],
},
}

-- Perform basic preprocessing on all events
for eventid, event in pairs(eventdefs) do
	-- Remove trailing tabs from quest
	if event.quest then
		event.quest = $:gsub("\t", "")
	end

	-- Remove trailing tabs from message
	for lineid, _ in ipairs(event) do
		if type(event[lineid]) == "string" then
			event[lineid] = $:gsub("\t", "")
		end
	end

	if eventid < 1001 then
		event.story = $ or eventid
	end

	-- Ensure all story conditions are stored as ranges
	if type(event.story) == "number" then
		event.story = {$, $}
	end
end
