/*
Battlefield Zone
	by Lach
	
SPOILER WARNING IF YOU HAVEN'T PLAYED THE LEVEL YET

If you want, you can check out the original Minecraft 1.12.2 superflat map this level is based on: https://cdn.discordapp.com/attachments/453514913760018435/766949570244116490/2020-10-15_21.25.26.png
I copied it as closely as I could entirely by hand :D
*/

// To prevent duplicate freeslots
local function SafeFreeslot(...)
	for _, item in ipairs({...})
		if rawget(_G, item) == nil
			freeslot(item)
		end
	end
end

SafeFreeslot("SPR_MCIT", "SPR_MCVI", "SPR_MCPT",
	"MT_MC_CROP", "MT_MC_ITEM", "MT_BTF_OPPONENT",
	"MT_MC_EOETARGET", "MT_MC_NPC", "MT_MC_NPC_WAYPOINT",
	"MT_MC_ALAVA", "MT_MC_LAVAPOP",
	"S_MC_CROP", "S_MC_LAVAPOP", "S_MC_SMOKE",
	"sfx_mcpop", "sfx_mcpois", "sfx_mcppon", "sfx_ygtmc", "sfx_ssbyee", "sfx_mcendp", "sfx_mcbust",
	"sfx_mclava", "sfx_mclavp")

for i = 1, 4
	SafeFreeslot("sfx_mcgrs"..i)
	sfxinfo[sfx_mcgrs1 + i - 1].caption = "Trample"
end

for i = 1, 4
	SafeFreeslot("sfx_mcldr"..i)
	sfxinfo[sfx_mcldr1 + i - 1].caption = "Ladder"
end

for i = 1, 2
	SafeFreeslot("sfx_mceoe"..i)
	sfxinfo[sfx_mceoe1 + i - 1].caption = "Throw"
end

for i = 1, 3
	SafeFreeslot("sfx_mceye"..i)
	sfxinfo[sfx_mceye1 + i - 1].caption = "Eye of Ender placed"
end

for i = 1, 3
	SafeFreeslot("sfx_mcvil"..i)
	sfxinfo[sfx_mcvil1 + i - 1].caption = "Mumble"
end

sfxinfo[sfx_mcpop].caption = "Plop"
sfxinfo[sfx_mcpois].caption = "Ew!"
sfxinfo[sfx_mcppon].caption = "Click"
sfxinfo[sfx_ygtmc].caption = "\"You're going to Minecraft!\""
sfxinfo[sfx_ssbyee].caption = "PING!"
sfxinfo[sfx_mcendp].caption = "End Portal activated"
sfxinfo[sfx_mclava].caption = "/" // [Stream] annoys the SHIT out of me
sfxinfo[sfx_mclavp].caption = "Lava pops"
sfxinfo[sfx_mcbust].caption = "/"

function A_MinecraftCrop(mo)
	if not mo.health
		mo.tics = 1
	end
end

function A_MinecraftSmoke(mo)
	local radius = FixedDiv(mo.radius, mo.scale)
	local height = FixedDiv(mo.height, mo.scale)
	local part = P_SpawnMobjFromMobj(mo,
		FixedMul(2*radius, P_RandomFixed()) - radius,
		FixedMul(2*radius, P_RandomFixed()) - radius,
		FixedMul(height, P_RandomFixed()), MT_PARTICLE)
	part.state = S_MC_SMOKE
	part.color = P_RandomChance(FRACUNIT/2) and SKINCOLOR_CARBON or SKINCOLOR_BLACK
	P_SetObjectMomZ(part, FRACUNIT >> 1)
end

states[S_MC_CROP] = {
	sprite = SPR_MCIT,
	tics = -1,
	action = A_MinecraftCrop
}

states[S_MC_LAVAPOP] = {
	sprite = SPR_MCPT,
	frame = FF_FULLBRIGHT,
	tics = 2,
	action = A_MinecraftSmoke,
	nextstate = S_MC_LAVAPOP
}

states[S_MC_SMOKE] = {
	sprite = SPR_MCPT,
	frame = B|FF_ANIMATE,
	tics = 64,
	var1 = 7,
	var2 = 8
}

mobjinfo[MT_MC_CROP] = {
	--$Name Minecraft Crop
	--$Sprite MCITH0
	--$Category Battlefield
	doomednum = 1540,
	spawnstate = S_MC_CROP,
	deathstate = S_MC_CROP,
	radius = 8*FRACUNIT,
	height = 16*FRACUNIT,
	flags = MF_SCENERY|MF_NOCLIP|MF_NOCLIPTHING
}

mobjinfo[MT_MC_ITEM] = {
	--$Name Eye of Ender
	--$Sprite MCITVLVR
	--$Category Battlefield
	doomednum = 1541,
	spawnstate = S_INVISIBLE,
	radius = 8*FRACUNIT,
	height = 16*FRACUNIT,
	painchance = ANG2,
	reactiontime = 24,
	speed = 4*FRACUNIT,
	flags = MF_SCENERY|MF_SPECIAL
}

mobjinfo[MT_BTF_OPPONENT] = {
	--$Name Battlefield Opponent
	--$Category Battlefield
	doomednum = 1542,
	spawnstate = S_INVISIBLE,
	radius = mobjinfo[MT_PLAYER].radius,
	height = mobjinfo[MT_PLAYER].height,
	reactiontime = 2*TICRATE
}

mobjinfo[MT_MC_EOETARGET] = {
	--$Name End Portal Frame Slot
	--$Category Battlefield
	doomednum = 1543,
	spawnstate = S_INVISIBLE,
	radius = 16*FRACUNIT,
	height = 32*FRACUNIT,
	flags = MF_SCENERY|MF_NOCLIP|MF_NOCLIPTHING|MF_NOBLOCKMAP|MF_NOGRAVITY
}

mobjinfo[MT_MC_NPC] = {
	--$Name Minecraft Villager
	--$Sprite MCVIAR
	--$Category Battlefield
	doomednum = 1544,
	spawnstate = S_INVISIBLE,
	radius = 5*FRACUNIT,
	height = 34*FRACUNIT,
	flags = MF_NOCLIP|MF_SCENERY,
	reactiontime = 10,
	speed = 8*FRACUNIT
}

mobjinfo[MT_MC_NPC_WAYPOINT] = {
	--$Name Minecraft Villager Waypoint
	--$Category Battlefield
	doomednum = 1545,
	spawnstate = S_INVISIBLE,
	radius = 8*FRACUNIT,
	height = 16*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPTHING|MF_SCENERY|MF_NOGRAVITY
}

mobjinfo[MT_MC_ALAVA] = {
	--$Name Minecraft Lava Ambience
	--$Sprite MCPTA0
	--$Category Battlefield
	doomednum = 1546,
	spawnstate = S_INVISIBLE,
	radius = 16*FRACUNIT,
	height = 16*FRACUNIT,
	spawnhealth = 7*TICRATE+TICRATE/2,
	seesound = sfx_mclava,
	flags = MF_NOBLOCKMAP|MF_NOSECTOR|MF_NOGRAVITY|MF_AMBIENT
}

mobjinfo[MT_MC_LAVAPOP] = {
	spawnstate = S_MC_LAVAPOP,
	seesound = sfx_mclavp,
	painchance = 2*FRACUNIT,
	radius = 4*FRACUNIT,
	height = 2*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_NOCLIPTHING
}

local function valid(mo)
	return mo and mo.valid
end

local HUD_NONE = 0
local HUD_SMASH = 1
local HUD_MINECRAFT = 2

local SQUISH_FACTOR = FRACUNIT/2
local SQUISH_TICS = 8

local THROW_DELAY = 14

local LAVAPOP_MINFUSE = TICRATE/2
local LAVAPOP_MAXFUSE = 4*TICRATE

local DEFAULT_DRAWDIST = 8192*FRACUNIT // if "drawdist" is set to "Infinite" then use this value
local CROP_DRAWFACTOR = FRACUNIT/5 // fraction of "drawdist" within which crops are rendered simplistically
local CROP_SIMPFACTOR = FRACUNIT/12 // fraction of "drawdist" within which crops are rendererd with detail
local CROP_ADDSEARCHRANGE = 64*FRACUNIT // extra blockmap range, so that crops outside the radius can be made invisible

local CROP_WHEAT = 0
local CROP_CARROTS = 1
local CROP_BEETROOT = 2
local CROP_POTATOES = 3

local CROP_STAGES = 8
local CROP_MINFUSE, CROP_MAXFUSE = 30*TICRATE, 90*TICRATE

local CROP_FRAMES = {
	[CROP_WHEAT] = {A, B, C, D, E, F, G, H},
	[CROP_CARROTS] = {I, I, J, J, K, K, K, L},
	[CROP_BEETROOT] = {I, I, J, J, K, K, K, M},
	[CROP_POTATOES] = {I, I, J, J, K, K, K, N}
}

local CROP_LOOT = {
	[CROP_WHEAT] = {O, P},
	[CROP_CARROTS] = {Q, Q},
	[CROP_BEETROOT] = {R, S},
	[CROP_POTATOES] = {T, T}
}

local CROP_SCORES = {
	[O] = 10,
	[P] = 100,
	[Q] = 50,
	[R] = 10,
	[S] = 100,
	[T] = 50,
	[U] = -5000
}

local EOE_FRAME = V

local NPC_SCALE = 3725*FRACUNIT/1000
local NPC_COLORS = {SKINCOLOR_FOREST, SKINCOLOR_PLUM, SKINCOLOR_BONE, SKINCOLOR_BROWN}
local NPC_MINFUSE, NPC_MAXFUSE = 3*TICRATE, 15*TICRATE

local NPC_PARTS = {
	// feet
	{
		frame = I, // front
		angle = 0,
		x = 0,
		y = 2,
		z = 0
	},
	{
		frame = J, // left
		angle = ANGLE_90,
		x = -3,
		y = 0,
		z = 0
	},
	{
		frame = K, // right
		angle = -ANGLE_90,
		x = 3,
		y = 0,
		z = 0
	},
	{
		frame = L, // back
		angle = ANGLE_180,
		x = 0,
		y = -2,
		z = 0
	},
	// body
	{
		frame = E, // front
		angle = 0,
		x = 0,
		y = 3,
		z = 6
	},
	{
		frame = F, // left
		angle = ANGLE_90,
		x = -4,
		y = 0,
		z = 6
	},
	{
		frame = G, // right
		angle = -ANGLE_90,
		x = 4,
		y = 0,
		z = 6
	},
	{
		frame = H, // back
		angle = ANGLE_180,
		x = 0,
		y = -3,
		z = 6
	},
	// arms
	{
		frame = M, // front
		angle = 0,
		x = 0,
		y = 7,
		z = 18
	},
	{
		frame = O, // front (inside)
		angle = ANGLE_180,
		x = 0,
		y = 4,
		z = 18
	},
	{
		frame = N, // left
		angle = ANGLE_90,
		x = -8,
		y = 3,
		z = 18
	},
	{
		frame = Q, // left (inside)
		angle = -ANGLE_90,
		x = -4,
		y = 3,
		z = 18
	},
	{
		frame = N, // right
		angle = -ANGLE_90,
		x = 8,
		y = 3,
		z = 18
	},
	{
		frame = Q, // right (inside)
		angle = ANGLE_90,
		x = 4,
		y = 3,
		z = 18
	},
	{
		frame = P, // back (left)
		angle = ANGLE_180,
		x = -6,
		y = -1,
		z = 18
	},
	{
		frame = P, // back (right)
		angle = ANGLE_180,
		x = 6,
		y = -1,
		z = 18
	},
	// head
	{
		frame = A, // front (face)
		angle = 0,
		x = 0,
		y = 4,
		z = 24
	},
	{
		frame = B, // left
		angle = ANGLE_90,
		x = -4,
		y = 0,
		z = 24
	},
	{
		frame = B, // right
		angle = -ANGLE_90,
		x = 4,
		y = 0,
		z = 24
	},
	{
		frame = B, // back
		angle = ANGLE_180,
		x = 0,
		y = -4,
		z = 24
	},
	// nose
	{
		frame = C, // front
		angle = 0,
		x = 0,
		y = 6,
		z = 23
	},
	{
		frame = D, // left
		angle = ANGLE_90,
		x = -1,
		y = 5,
		z = 23
	},
	{
		frame = D, // right
		angle = -ANGLE_90,
		x = 1,
		y = 5,
		z = 23
	}
}

local function BattlefieldZone()
	return (mapheaderinfo[gamemap] and mapheaderinfo[gamemap].btfmczone)
end

// local vars
local cv_drawdist = CV_FindVar("drawdist")
local btfmc_hud = HUD_NONE
local clientplayers
local smashplayers
local hidecrops = false

// netvars
local eyes = 0
local mapeyes
local waypoints
local opponent
local playerdamage = 0

addHook("NetVars", function(net)
	eyes = net(eyes)
	mapeyes = net(mapeyes)
	waypoints = net(waypoints)
	opponent = net(opponent)
	playerdamage = net(playerdamage)
end)

addHook("MapChange", function()
	eyes = 0
	waypoints = nil
	clientplayers = nil
	smashplayers = nil
	opponent = nil
	playerdamage = 0
	if btfmc_hud ~= HUD_NONE
		btfmc_hud = HUD_NONE
		hud.enable("lives")
	end
	for player in players.iterate
		player.btfmc_inventory = nil
	end
end)

// Collectible items

local function EOESpawn(mo)
	mo.scale = $ << 1
	mo.sprite = SPR_MCIT
	mo.frame = EOE_FRAME | FF_PAPERSPRITE | FF_FULLBRIGHT
	mo.flags = $ | MF_NOGRAVITY
	mo.flags2 = $ | MF2_INFLOAT // designates this item as an eye of ender
	if not (mo.spawnpoint and mo.spawnpoint.options & MTF_AMBUSH)
		mo.z = $ + P_MobjFlip(mo)*16*FRACUNIT
	end
end

local function CollectEOE(mo)
	eyes = $ + 1
	for i, eye in ipairs(mapeyes)
		if eye == mo
			table.remove(mapeyes, i)
			return
		end
	end
end

local function CreateInventoryEntry()
	return {
		count = 0,
		frame = A,
		squishtics = 0,
		throwdelay = 0
	}
end

local function GetItemEntry(inventory, frame)
	for i, entry in ipairs(inventory)
		if entry.frame == frame
		and entry.count > 0
			return entry
		end
	end
end

local function CollectItem(player, frame, count, sound)
	local inventory
	local slots
	local found = false
	
	local function AddToEntry(entry)
		entry.frame = frame
		entry.count = $ + (count or 1)
		entry.squishtics = SQUISH_TICS
		S_StartSound(player.mo, sound or sfx_mcpop)
		return entry
	end
	
	// first, check that the player has an inventory. if not, create one, and add the item to the first slot
	if not player.btfmc_inventory
		inventory = {}
		table.insert(inventory, AddToEntry(CreateInventoryEntry()))
		player.btfmc_inventory = inventory
		return
	end
	
	inventory = player.btfmc_inventory
	slots = #player.btfmc_inventory
	
	// next, check to see if the player already has a slot containing this type of item, and use that slot if so
	for i = 1, slots
		local entry = inventory[i]
		if entry.frame == frame and entry.count < 64 and entry.count > 0
			AddToEntry(entry)
			return
		end
	end
	
	// failing that, check whether any existing slots are empty
	for i = 1, slots
		local entry = inventory[i]
		if entry.count == 0
			AddToEntry(entry)
			return
		end
	end
	
	// failing that, create a new inventory slot and add the item there
	table.insert(inventory, AddToEntry(CreateInventoryEntry()))
end

local function TryThrowItem(player, frame, spawn)
	local inventory = player.btfmc_inventory
	local entry = inventory and GetItemEntry(inventory, frame)
	
	if not entry
	or entry.throwdelay
		return false
	end
	
	if spawn
		// TODO
	end
	entry.count = $ - 1
	entry.throwdelay = THROW_DELAY
	return true
end

addHook("MapThingSpawn", EOESpawn, MT_MC_ITEM)

addHook("MobjThinker", function(mo)
	mo.angle = $ + mo.info.painchance
	
	if mo.reactiontime
		mo.reactiontime = $ - 1
	end
	
	if mo.fuse and mo.fuse < 2*TICRATE
		mo.flags2 = $ ^^ MF2_DONTDRAW
	end
	
	if mo.flags2 & MF2_INFLOAT
	and leveltime & 1
		local radius = FixedDiv(mo.radius, mo.scale)
		local height = FixedDiv(mo.height, mo.scale)
		local spark = P_SpawnMobjFromMobj(mo,
			FixedMul(P_RandomFixed(), 2*radius) - radius,
			FixedMul(P_RandomFixed(), 2*radius) - radius,
			FixedMul(P_RandomFixed(), height),
			MT_PARTICLE)
		spark.scale = $ >> 1
		spark.state = S_NIGHTSPARKLE1
		spark.tics = states[S_BOXSPARKLE1].tics
		spark.color = SKINCOLOR_MAGENTA
		spark.colorized = true
		
		P_SetObjectMomZ(spark, FRACUNIT)
	end
	
	if mo.target
		local target = mo.target
		
		if not target.valid // ??
			P_RemoveMobj(mo)
			return
		end
		
		if not mo.reactiontime
			local moz = mo.z + (mo.height >> 1)
			local targetz = target.z + (target.height >> 1)
			local dist = FixedHypot(FixedHypot(mo.x - target.x, mo.y - target.y), moz - targetz)
			local speed
			
			if target.player
				mo.movefactor = $ + 2*mo.scale
				speed = mo.movefactor
			else
				speed = FixedMul(mo.info.speed, mo.scale)
			end
			if speed >= dist
				P_KillMobj(mo, target, target)
			else
				mo.momx = FixedMul(FixedDiv(target.x - mo.x, dist), speed)
				mo.momy = FixedMul(FixedDiv(target.y - mo.y, dist), speed)
				mo.momz = FixedMul(FixedDiv(targetz - moz, dist), speed)
			end
		end
	end
end, MT_MC_ITEM)

addHook("TouchSpecial", function(mo, pmo)
	local player = pmo.player
	
	if mo.reactiontime
	or valid(mo.target)
	or player.bot
		return true
	end
	
	mo.target = pmo
	mo.flags = $ | MF_NOGRAVITY
	mo.fuse = 0
	mo.flags2 = $ & ~MF2_DONTDRAW
	
	if mo.flags2 & MF2_INFLOAT
		CollectEOE(mo)
	end
	CollectItem(player, mo.frame & FF_FRAMEMASK, 1, mo.threshold or sfx_mcpop)
	
	return true
end, MT_MC_ITEM)

addHook("MobjDeath", function(mo, inflictor)
	if not valid(inflictor) then return end
	
	local player = inflictor.player
	local score = CROP_SCORES[mo.frame & FF_FRAMEMASK]
	
	if player
		if score
			P_AddPlayerScore(player, max(-player.score, score))
		end
	elseif mo.flags2 & MF2_INFLOAT
		assert(valid(inflictor.target), "\133A player respawned or left the game before a vital linedef executor could be triggered. Please reload the level, sorry. :(")
		P_LinedefExecute(1300, inflictor.target, inflictor.subsector.sector)
		S_StartSound(inflictor, sfx_mceye1 + P_RandomKey(3))
	end
end, MT_MC_ITEM)

// Tramplable crops

addHook("MapThingSpawn", function(mo, mt)
	local leaf = mo
	local ang = 0
	local frame
	
	mo.movedir = CROP_FRAMES[mt.angle] and mt.angle or CROP_WHEAT
	mo.movecount = P_RandomKey(CROP_STAGES) + 1
	mo.fuse = P_RandomRange(CROP_MINFUSE, CROP_MAXFUSE)
	mo.scale = 4*$
	mo.flags2 = $ | MF2_DONTDRAW | MF2_SLIDEPUSH
	
	frame = CROP_FRAMES[mo.movedir][mo.movecount]
	mo.frame = frame
	frame = $ | FF_PAPERSPRITE
	
	for i = 1, 4
		leaf.tracer = P_SpawnMobjFromMobj(mo,
			P_ReturnThrustX(mo, ang, 4*FRACUNIT),
			P_ReturnThrustY(mo, ang, 4*FRACUNIT), 0, MT_THOK)
		leaf = leaf.tracer
		leaf.flags2 = $ | MF2_DONTDRAW
		leaf.state = mo.state
		leaf.frame = frame
		leaf.angle = ang + ANGLE_90
		ang = $ + ANGLE_90
	end
end, MT_MC_CROP)

local crop_drawdist = 0
local crop_simpdist = 0
local function CropSearch(pmo, mo)
	if not (mo.valid and mo.type == MT_MC_CROP)
		return
	end
	
	local closestdist = INT32_MAX
	
	for i, pmo in ipairs(clientplayers)
		local dist = FixedHypot(pmo.x - mo.x, pmo.y - mo.y)
		if dist < closestdist
			closestdist = dist
		end
	end

	if closestdist > crop_drawdist // don't draw anything
		if mo.flags2 & MF2_SLIDEPUSH
			local leaf = mo.tracer
			while leaf
				leaf.flags2 = $ | MF2_DONTDRAW
				leaf = leaf.tracer
			end
			mo.flags2 = $ & ~MF2_SLIDEPUSH
		end
		mo.flags2 = $ | MF2_DONTDRAW
	elseif closestdist > crop_simpdist // draw one sprite
		if mo.flags2 & MF2_SLIDEPUSH
			local leaf = mo.tracer
			while leaf
				leaf.flags2 = $ | MF2_DONTDRAW
				leaf = leaf.tracer
			end
			mo.flags2 = $ & ~MF2_SLIDEPUSH
		end
		mo.flags2 = $ & ~MF2_DONTDRAW
	else // draw papersprites
		if not (mo.flags2 & MF2_SLIDEPUSH)
			local leaf = mo.tracer
			while leaf
				leaf.flags2 = $ & ~MF2_DONTDRAW
				leaf = leaf.tracer
			end
			mo.flags2 = $ | MF2_SLIDEPUSH
		end
		mo.flags2 = $ | MF2_DONTDRAW
	end
end

addHook("ThinkFrame", do
	if not BattlefieldZone() then return end
	
	// hide all crops on viewswitch
	if hidecrops
		hidecrops = false
		for mt in mapthings.iterate
			if mt.type == mobjinfo[MT_MC_CROP].doomednum
			and valid(mt.mobj)
				local mo = mt.mobj
				if mo.flags2 & MF2_SLIDEPUSH
					local leaf = mo.tracer
					while leaf
						leaf.flags2 = $ | MF2_DONTDRAW
						leaf = leaf.tracer
					end
					mo.flags2 = $ & ~MF2_SLIDEPUSH
				end
				mo.flags2 = $ | MF2_DONTDRAW
			end
		end
	end
	
	// crop draw distance
	if cv_drawdist
		local minx = INT32_MAX
		local maxx = INT32_MIN
		local miny = INT32_MAX
		local maxy = INT32_MIN
		
		clientplayers = {}
		table.insert(clientplayers, displayplayer and displayplayer.realmo)
		table.insert(clientplayers, secondarydisplayplayer and secondarydisplayplayer.realmo)
		
		if #clientplayers
			local drawdist = cv_drawdist.value*FRACUNIT or DEFAULT_DRAWDIST
			local searchdist
			
			crop_drawdist = FixedMul(drawdist, CROP_DRAWFACTOR)
			crop_simpdist = FixedMul(drawdist, CROP_SIMPFACTOR)
			searchdist = crop_drawdist + CROP_ADDSEARCHRANGE
			
			for i, mo in ipairs(clientplayers)
				minx = min($, mo.x)
				maxx = max($, mo.x)
				miny = min($, mo.y)
				maxy = max($, mo.y)
			end
			
			searchBlockmap("objects", CropSearch, clientplayers[1],
				minx - searchdist, maxx + searchdist,
				miny - searchdist, maxy + searchdist)
		end
	end
	
	// inventory ticker
	local numeyes = 0
	for player in players.iterate
		local inventory = player.btfmc_inventory
		
		if not inventory then continue end
		
		for i, entry in ipairs(inventory)
			if entry.frame == EOE_FRAME
				numeyes = $ + entry.count
			end
			if entry.throwdelay
				entry.throwdelay = $ - 1
			end
		end
	end
	
	while numeyes < eyes // a player may have left the game? give the deleted eyes to other players
		for player in players.iterate
			local inventory = player.btfmc_inventory
			
			if not inventory and player.mo then continue end
			
			CollectItem(player, EOE_FRAME)
			numeyes = $ + 1
			if numeyes == eyes then break end
		end
	end	
end)

addHook("ViewpointSwitch", function()
	if not BattlefieldZone() then return end
	hidecrops = true
end)

addHook("MobjFuse", function(mo)
	if mo.movecount < CROP_STAGES
		local frame
		local leaf = mo.tracer
		
		mo.movecount = $ + 1
		mo.fuse = P_RandomRange(CROP_MINFUSE, CROP_MAXFUSE)
		
		frame = CROP_FRAMES[mo.movedir][mo.movecount]
		mo.frame = frame
		frame = $ | FF_PAPERSPRITE
		
		while leaf
			leaf.frame = frame
			leaf = leaf.tracer
		end
	end
	return true
end, MT_MC_CROP)

addHook("MobjDeath", function(mo, inflictor)
	local function SpawnItem(frame)
		local item = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_MC_ITEM)
		item.scale = $ >> 1
		item.angle = P_RandomKey(360)*ANG1
		P_InstaThrust(item, P_RandomKey(360)*ANG1, item.scale >> 1)
		P_SetObjectMomZ(item, 2*item.scale)
		item.sprite = SPR_MCIT
		item.frame = frame | FF_PAPERSPRITE
		item.fuse = 6*TICRATE
		return item
	end
	
	local leaf = mo.tracer
	
	while valid(leaf)
		local temp = leaf.tracer
		P_RemoveMobj(leaf)
		leaf = temp
	end
	
	if mo.movecount < CROP_STAGES
		SpawnItem(CROP_LOOT[mo.movedir][1])
	else
		SpawnItem(CROP_LOOT[mo.movedir][2])
		for i = 1, 3
			if P_RandomChance(57*FRACUNIT/100)
				SpawnItem(CROP_LOOT[mo.movedir][1])
			end
		end
		if mo.movedir == CROP_POTATOES
		and P_RandomChance(2*FRACUNIT/100)
			SpawnItem(CROP_LOOT[mo.movedir][2] + 1).threshold = sfx_mcpois
		end
	end
	
	if valid(inflictor)
		S_StartSound(inflictor, sfx_mcgrs1 + P_RandomKey(4))
	end
end, MT_MC_CROP)

// Battlefield Sonic

addHook("MobjSpawn", function(mo)
	mo.skin = "sonic"
	if skins[mo.skin].realname:lower() ~= mo.skin // skin patch catch - if playing as Ponic, Sonic should still be blue
		mo.flags2 = $ | MF2_DONTRESPAWN // don't bother colorizing
		mo.color = SKINCOLOR_BLUE
	else
		mo.color = skins[mo.skin].prefcolor
	end
	mo.state = S_PLAY_STND
	mo.shadowscale = FRACUNIT
end, MT_BTF_OPPONENT)

addHook("MobjThinker", function(mo)
	if not (mo.flags2 & (MF2_DONTRESPAWN|MF2_SLIDEPUSH))
		mo.flags2 = $ | MF2_DONTRESPAWN
		for player in players.iterate
			if player.mo and player.mo.skin == mo.skin
				mo.color = SKINCOLOR_SILVER
				mo.colorized = true
				mo.flags2 = $ | MF2_SLIDEPUSH
				break
			end
		end
	end
	
	if mo.flags2 & MF2_SUPERFIRE
		return
	end
	
	if mo.reactiontime
		mo.friction = FRACUNIT
		mo.reactiontime = $ - 1
		return
	end
	
	if mo.state == S_PLAY_STND
		if mo.flags2 & MF2_FIRING
			mo.state = S_PLAY_GASP
			P_SetObjectMomZ(mo, 4*mo.scale)
			S_StartSound(mo, sfx_ssbyee)
			for player in players.iterate
				local pmo = player.mo
				if pmo
					player.powers[pw_nocontrol] = TICRATE
					pmo.state = S_PLAY_PAIN
					pmo.z = $ + P_MobjFlip(mo)
					P_SetObjectMomZ(pmo, 7*FRACUNIT)
					P_InstaThrust(pmo, mo.angle, 60*mo.scale)
				end
			end
			playerdamage = 999 // LOL
		elseif mo.flags2 & MF2_JUSTATTACKED
			S_StopMusic()
			for player in players.iterate
				P_FlashPal(player, PAL_WHITE, 3)
			end
			S_StartSound(mo, sfx_ygtmc)
			mo.flags2 = $ | MF2_FIRING
			mo.reactiontime = 2*TICRATE
		else
			mo.state = S_PLAY_WALK
			P_InstaThrust(mo, mo.angle, 4*mo.scale)
			mo.reactiontime = 5*TICRATE/2
		end
		return
	end
	
	if mo.state == S_PLAY_WALK
		if mo.flags2 & MF2_FIRING
			mo.state = S_PLAY_STND
			mo.flags2 = $ | MF2_SUPERFIRE
		elseif FixedHypot(mo.momx, mo.momy) < mo.scale
			mo.momx = 0
			mo.momy = 0
			mo.state = S_PLAY_STND
			mo.tics = -1
			mo.flags2 = $ | MF2_JUSTATTACKED
			mo.reactiontime = TICRATE/4
		end
		return
	end
end, MT_BTF_OPPONENT)

// Villager

local function CombineTables(t1, t2)
	local ret = {}
	for i, t in ipairs({t1, t2})
		for k, v in ipairs(t)
			table.insert(ret, v)
		end
	end
	return ret
end

local VF_DONTSEARCHHNEXT = 1 << 0
local VF_DONTSEARCHTRACER = 1 << 1
local VF_DONTSEARCHHPREV = 1 << 2
local function GetWaypointPath(current, target, flags)
	if current == nil // hit the end of the path, target waypoint is not here :(
		return nil
	end
	
	if current == target // we found the target! :D let's send a list back to store the other waypoints and create a path
		return {target}
	end
	
	flags = $ or 0
	
	local path = ~(flags & VF_DONTSEARCHHNEXT) and GetWaypointPath(current.hnext, target, VF_DONTSEARCHHPREV)
	or ~(flags & VF_DONTSEARCHTRACER) and GetWaypointPath(current.tracer, target, VF_DONTSEARCHHPREV)
	
	if not path // target waypoint was not found in any forward-searched path
		if flags & VF_DONTSEARCHHPREV
			return nil
		end
		
		local list = {}
		local branch
		repeat
			table.insert(list, 1, current)
			branch = current
			current = current.hprev
		until current.tracer // add backwards waypoints to the path until we find a waypoint with a branch
		return CombineTables(GetWaypointPath(current, target, current.tracer == branch and VF_DONTSEARCHTRACER or VF_DONTSEARCHHNEXT), list)
	end
	
	// attach the current waypoint to the returned path of waypoints
	table.insert(path, current)
	return path
end

local function UpdateVillagerParts(mo)
	local forwardx = FixedMul(cos(mo.angle), mo.scale)
	local forwardy = FixedMul(sin(mo.angle), mo.scale)
	local sidex = FixedMul(cos(mo.angle - ANGLE_90), mo.scale)
	local sidey = FixedMul(sin(mo.angle - ANGLE_90), mo.scale)
	
	for i, part in ipairs(mo.parts)
		local info = NPC_PARTS[i]
		P_TeleportMove(part,
			mo.x + info.y*forwardx + info.x*sidex,
			mo.y + info.y*forwardy + info.x*sidey,
			mo.z + info.z*mo.scale)
		part.angle = part.movedir + mo.angle
	end
end

addHook("MapThingSpawn", function(mo, mt)
	if mt.options & MTF_AMBUSH // blacksmith (sort of)
		mo.flags2 = $ & MF2_AMBUSH
		mo.color = SKINCOLOR_BLACK
	else // random cloak color
		mo.color = NPC_COLORS[P_RandomKey(#NPC_COLORS) + 1]
	end
	
	// initialize papersprite parts
	mo.scale = FixedMul($, NPC_SCALE)
	mo.parts = {}
	for i, partinfo in ipairs(NPC_PARTS)
		local part = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
		part.state = S_INVISIBLE
		part.sprite = SPR_MCVI
		part.color = mo.color
		part.frame = partinfo.frame | FF_PAPERSPRITE
		part.movedir = partinfo.angle + ANGLE_90
		table.insert(mo.parts, part)
	end
	UpdateVillagerParts(mo)
end, MT_MC_NPC)

addHook("MobjThinker", function(mo)
	if mo.reactiontime // just spawned? delay a little before setup
		mo.reactiontime = $ - 1
		if not mo.reactiontime
			local bestpoint
			local bestdist = INT32_MAX
			
			// search for closest waypoint to be initial tracer
			for i, waypoint in ipairs(waypoints)
				local dist
				waypoint = waypoint.mobj
				dist = FixedHypot(mo.x - waypoint.x, mo.y - waypoint.y)
				if dist < bestdist
					bestpoint = waypoint
					bestdist = dist
				end
			end
			mo.tracer = bestpoint
			mo.target = bestpoint
			mo.angle = R_PointToAngle2(mo.x, mo.y, mo.tracer.x, mo.tracer.y)
			
			UpdateVillagerParts(mo)
		end
	elseif mo.fuse // play sounds
		if (mo.fuse % (2*TICRATE)) == 0
			if P_RandomChance(FRACUNIT/4)
				S_StartSound(mo, sfx_mcvil1 + P_RandomKey(3))
			end
		end
	else // movement
		local tracer = mo.tracer
		local target = mo.target
		local dist = FixedHypot(mo.x - tracer.x, mo.y - tracer.y)
		
		if dist < mo.info.speed // close enough to waypoint
			mo.momx = 0
			mo.momy = 0
			P_TeleportMove(mo, tracer.x, tracer.y, tracer.floorz)
			
			if tracer == target // reached our destination!
				mo.fuse = P_RandomRange(NPC_MINFUSE, NPC_MAXFUSE)
				UpdateVillagerParts(mo)
				return
			end
			
			// set next waypoint
			mo.tracer = table.remove(mo.path)
			mo.angle = R_PointToAngle2(mo.x, mo.y, mo.tracer.x, mo.tracer.y)
		end
		
		// move towards waypoint
		P_InstaThrust(mo, mo.angle, mo.info.speed)
		UpdateVillagerParts(mo)
	end
end, MT_MC_NPC)

addHook("MobjFuse", function(mo)
	local validpoints = {}
	local newtarget
	
	// search for all unoccupied waypoints
	for i, waypoint in ipairs(waypoints)
		waypoint = waypoint.mobj
		if not waypoint.target
			table.insert(validpoints, waypoint)
		end
	end
	
	// all waypoints occupied? shouldn't happen, but try again later
	if not #validpoints
		print("Villager found no target waypoints!")
		mo.fuse = P_RandomRange(NPC_MINFUSE, NPC_MAXFUSE)
		return true
	end
	
	newtarget = validpoints[P_RandomKey(#validpoints) + 1] // choose random unoccupied waypoint
	
	mo.path = GetWaypointPath(mo.target, newtarget)
	table.remove(mo.path) // the first waypoint in the path is always the waypoint the villager is at, so don't worry about travelling there.
	
	mo.target.target = nil // free the current waypoint's target so other villagers can go there
	mo.target = newtarget
	mo.target.target = mo // set the new waypoint target to the villager; now no other villager should occupy this spot
	return true
end, MT_MC_NPC)

// Lava pops

addHook("MapThingSpawn", function(mo, mt)
	local particleradius = FixedMul(mobjinfo[MT_MC_LAVAPOP].radius, mobjinfo[MT_MC_LAVAPOP].painchance)
	local sector = mo.subsector.sector
	local vertices = {}
	local leftbound = INT32_MIN
	local rightbound = INT32_MAX
	local topbound = INT32_MAX
	local bottombound = INT32_MIN
	
	// get a list of vertices that make up the sector
	for i = 0, #sector.lines - 1
		local line = sector.lines[i]
		local ends = {line.v1, line.v2}
		
		for j, v1 in ipairs(ends)
			local haveit = false
			for k, v2 in ipairs(vertices)
				if v1 == v2
					haveit = true
					break
				end
			end
			if not haveit
				table.insert(vertices, v1)
			end
		end
	end
	
	// check each vertex and create a bounding box of sector area surrounding the spawner
	// (we can assume a bounding box is a safe spawn area because this is fucking minecraft)
	for i, vertex in ipairs(vertices)
		if vertex.x < rightbound
		and vertex.x >= mo.x
			rightbound = vertex.x
		elseif vertex.x > leftbound
		and vertex.x <= mo.x
			leftbound = vertex.x
		end
		
		if vertex.y < topbound
		and vertex.y >= mo.y
			topbound = vertex.y
		elseif vertex.y > bottombound
		and vertex.y <= mo.y
			bottombound = vertex.y
		end
	end
	
	// reduce bounding zones by size of particles
	leftbound = $ + particleradius
	rightbound = $ - particleradius
	topbound = $ - particleradius
	bottombound = $ + particleradius
	
	// is there still enough space to spawn particles?
	if rightbound - leftbound < 0
	or topbound - bottombound < 0
		return
	end
	
	// set bounds and fuse!
	mo.leftbound = leftbound
	mo.rightbound = rightbound
	mo.topbound = topbound
	mo.bottombound = bottombound
	mo.fuse = P_RandomRange(LAVAPOP_MINFUSE, LAVAPOP_MAXFUSE)
end, MT_MC_ALAVA)

addHook("MobjThinker", function(mo)
	if mo.fuse
		mo.fuse = $ - 1
		if mo.fuse == 0
			local part = P_SpawnMobj(mo.leftbound + FixedMul(mo.rightbound - mo.leftbound, P_RandomFixed()),
				mo.bottombound + FixedMul(mo.topbound - mo.bottombound, P_RandomFixed()),
				mo.z, MT_MC_LAVAPOP)
			part.scale = FixedMul($, part.info.painchance)
			S_StartSound(part, part.info.seesound)
			P_InstaThrust(part, P_RandomKey(360)*ANG1, 2*FRACUNIT + 3*P_RandomFixed())
			P_SetObjectMomZ(part, 4*FRACUNIT + 3*P_RandomFixed())
			mo.fuse = P_RandomRange(LAVAPOP_MINFUSE, LAVAPOP_MAXFUSE)
		end
	end
end, MT_MC_ALAVA)

addHook("MobjThinker", function(mo)
	if P_MobjFlip(mo)*mo.momz <= 0
	and (P_IsObjectOnGround(mo) or mo.eflags & MFE_UNDERWATER)
		P_RemoveMobj(mo)
	end
end, MT_MC_LAVAPOP)

// Linedef executors

addHook("LinedefExecute", function(line, mo)
	local player = valid(mo) and mo.player
	
	if not player
	or not (mo.eflags & MFE_JUSTHITFLOOR
		or player.pflags & PF_GLIDING
		or player.charability2 == CA2_MELEE and player.panim == PA_ABILITY2
		or player.dashmode >= DASHMODE_THRESHOLD and player.charflags & SF_MACHINE)
			return
	end
	
	local sector = mo.subsector.sector
	local goalheight = line.frontside.textureoffset
	
	if sector.floorheight ~= goalheight
		sector.floorheight = goalheight
		sector.floorpic = line.frontsector.floorpic
		
		for m in sector.thinglist()
			if m.valid and m.type == MT_MC_CROP
				P_KillMobj(m, mo, mo)
			end
		end
	end
end, "FARMLAND")

addHook("LinedefExecute", function()
	local cam
	local numeyes = 0
	local numnpcs = 0
	local indexedwaypoints = {}
	waypoints = {}
	mapeyes = {}
	
	// this gets in the way of our custom huds, and this pack is probably going to get an inflives script anyway
	hud.disable("lives")
	btfmc_hud = HUD_SMASH
	
	// collect information on mapthings
	for mt in mapthings.iterate
		if mt.type == mobjinfo[MT_MC_ITEM].doomednum
			numeyes = $ + 1
			table.insert(mapeyes, mt.mobj)
		elseif mt.type == mobjinfo[MT_MC_NPC_WAYPOINT].doomednum
			if not indexedwaypoints[mt.angle]
				indexedwaypoints[mt.angle] = {}
			end
			table.insert(indexedwaypoints[mt.angle], mt.mobj)
			table.insert(waypoints, mt)
		elseif mt.type == mobjinfo[MT_MC_NPC].doomednum
			numnpcs = $ + 1
		elseif mt.type == mobjinfo[MT_BTF_OPPONENT].doomednum
			opponent = mt.mobj
		elseif mt.type == mobjinfo[MT_ALTVIEWMAN].doomednum
		and valid(mt.mobj)
			cam = mt.mobj
		end
	end
	
	// set awayview camera and nocontrol for all players
	for player in players.iterate
		if player.mo
			player.powers[pw_nocontrol] = 10*TICRATE
			if cam
				player.awayviewmobj = cam
				player.awayviewtics = 60
				player.awayviewaiming = -ANG20
			end
		end
	end
	
	// connect waypoint linked lists
	// all waypoints have one preceding waypoint (hprev) and up to two "forward" waypoints (hnext and tracer)
	for i, mt in ipairs(waypoints)
		if mt.z == 0 then continue end
		
		local waypoint = mt.mobj
		local targets = indexedwaypoints[mt.z]
		
		assert(targets and targets[1], "\133A waypoint with ID "..mt.angle.." could not find any waypoint with ID "..mt.z.."!")
		
		waypoint.hnext = targets[1]
		waypoint.hnext.hprev = waypoint
		if targets[2]
			waypoint.tracer = targets[2]
			waypoint.tracer.hprev = waypoint
		end
	end
	
	// ensure all waypoints have a preceding waypoint (hprev)
	// if I'm not a complete idiot, this should mean that all of them are reachable!
	for i, mt in ipairs(waypoints)
		assert(mt.mobj.hprev, "\133A waypoint with ID "..mt.angle.." does not have a preceding waypoint!")
	end
	
	// ensure there are 12 eyes of ender, otherwise the level isn't beatable!
	assert(numeyes == 12, "\133"..numeyes.." Eyes of Ender found! There should be 12.")
	
	// ensure there are at least enough waypoints to house all villagers
	assert(numnpcs < #waypoints, "\133"..numnpcs.." Villagers cannot fit into "..#waypoints.." waypoints!")
end, "BTFSTART")

addHook("LinedefExecute", function()
	for player in players.iterate
		P_DoPlayerExit(player)
	end
end, "BTFEXIT")

addHook("LinedefExecute", function(line, mo)
	local player = mo.valid and mo.player
	local entry
	
	if not player
	or not eyes // no eyes collected? don't bother calculating anything!
		return
	end
	
	if TryThrowItem(player, EOE_FRAME)
		local ang = INT32_MAX
		local target
		
		for mt in mapthings.iterate
			if mt.type == mobjinfo[MT_MC_EOETARGET].doomednum
			and mt.angle < ang
			and valid(mt.mobj)
			and not (mt.mobj.flags2 & MF2_DONTRESPAWN)
				ang = mt.angle
				target = mt.mobj
			end
		end
		
		assert(target, "\133No more Eye of Ender targets found!")
		
		local eye = P_SpawnMobjFromMobj(mo, 0, 0, player.height, MT_MC_ITEM)
		EOESpawn(eye)
		P_SetObjectMomZ(eye, eye.info.speed)
		eye.target = target
		eye.target.flags2 = $ | MF2_DONTRESPAWN
		eye.target.target = mo
		S_StartSound(mo, sfx_mceoe1 + P_RandomKey(2))
		eyes = $ - 1
	end
end, "BTFEOE")

addHook("LinedefExecute", function(line, mo, sector)
	sector.floorheight = line.frontsector.floorheight
	sector.floorpic = line.frontsector.floorpic
	P_LinedefExecute(2000, mo, sector)
end, "BTFEYEIN")

addHook("LinedefExecute", function()
	btfmc_hud = HUD_MINECRAFT
end, "BTFHUD")

// Ladder climbing

local LADDERSPEED = 5*64*FRACUNIT/(2*TICRATE)
addHook("MobjLineCollide", function(mo, line)
	if not BattlefieldZone()
		return
	end
	
	local player = mo.valid and mo.player
	
	if not player return end
	
	if player.cmd.forwardmove or player.cmd.sidemove
		for fof in line.frontsector.ffloors()
			if mo.z < fof.topheight
			and mo.z + mo.height > fof.bottomheight
			and fof.master.frontside.midtexture == R_TextureNumForName("MCLADDER")
				P_SetObjectMomZ(mo, LADDERSPEED)
				if leveltime % 14 == 0
					S_StartSoundAtVolume(mo, sfx_mcldr1 + P_RandomKey(4), 128)
				end
			end
		end
	end
end, MT_PLAYER)

// HUD stuff

local SMASH_COLORS = {
	SKINCOLOR_RED,
	SKINCOLOR_BLUE,
	SKINCOLOR_YELLOW,
	SKINCOLOR_GREEN
}
local MAX_LIVES = 4
local DRAW_LIVES = 3
local LIVES_XSPACE = 18*FRACUNIT
local LIVES_XOFFSET = -MAX_LIVES*LIVES_XSPACE/2 + 10*FRACUNIT
local LIVES_YOFFSET = -40*FRACUNIT
local DAMAGE_XOFFSET = 36*FRACUNIT
local DAMAGE_YOFFSET = -20*FRACUNIT

local INVENTORY_SLOTS = 9
local COMPASS_PATCHES = 32
local COMPASS_OFFSET = FixedAngle(360*FRACUNIT/(COMPASS_PATCHES*2))
local COMPASS_ACCEL = ANG2
local COMPASS_DECEL = ANG1*9/2

local HOTBAR
local SLOTS
local OFFHAND
local OFFSLOT
local WHITE
local COMPASS = {}

local compasseyes = {}
local compassangles = {0, 0}
local compassspeeds = {0, 0}

local function GetHotbarCoords()
	return (320 - HOTBAR.width)*FRACUNIT/2, (200 - HOTBAR.height)*FRACUNIT
end

local function DrawMinecraftNumber(v, x, y, scale, font, text, flags, color, shadow)
	local colormap = v.getColormap(TC_DEFAULT, color)
	local shadowmap
	
	if shadow
		shadowmap = v.getColormap(TC_DEFAULT, SKINCOLOR_BLACK)
	end
	
	text = tostring(text)
	
	for i = text:len(), 1, -1
		local y = y
		local patch = v.cachePatch(font..text:sub(i, i))
		
		if not patch then continue end
		
		x = $ - patch.width*scale
		y = $ - patch.height*scale
		
		if shadow
			v.drawScaled(x + scale, y + scale, scale, patch, flags, shadowmap)
		end
		v.drawScaled(x, y, scale, patch, flags, colormap)
		
		x = $ - scale
	end
end

local function SmashHUD(v, player, cam)
	if player == secondarydisplayplayer // this is more fun drawn as a single entity
		return
	end
	
	// vars
	local space = 320*FRACUNIT
	local x = 0
	local y = 190*FRACUNIT
	local patch
	
	// set up the player info!
	if not smashplayers
		if not valid(opponent) // don't even bother
			return
		end
		smashplayers = {}
		table.insert(smashplayers, {
			skin = "sonic",
			color = opponent.color,
			translation = opponent.colorized and TC_RAINBOW or TC_DEFAULT
		})
		for player in players.iterate
			if player.mo
				table.insert(smashplayers, {
					skin = player.mo.skin,
					color = player.skincolor,
					translation = TC_DEFAULT
				})
			end
			if #smashplayers >= 4
				break
			end
		end
	end
	
	// draw player info
	space = $ / #smashplayers
	x = $ + space/2
	for i, info in ipairs(smashplayers)
		local livesmap = v.getColormap(info.translation, info.color)
		local damage = i ~= 1 and playerdamage or 0
		local damagestring = damage.."%"
		local damagex = x + DAMAGE_XOFFSET
		
		// draw "franchise" icon (we're just doing character signposts)
		patch = v.getSprite2Patch(info.skin, SPR2_SIGN)
		v.drawScaled(x, y, FRACUNIT, patch, V_SNAPTOBOTTOM|V_HUDTRANSHALF, v.getColormap(TC_BLINK, SMASH_COLORS[i]))
		
		// draw lives icons
		patch = v.getSprite2Patch(info.skin, SPR2_LIFE)
		for j = 1, DRAW_LIVES
			v.drawScaled(x + LIVES_XOFFSET + (j - 1)*LIVES_XSPACE, y + LIVES_YOFFSET, FRACUNIT,
				patch, V_SNAPTOBOTTOM|V_HUDTRANS, livesmap)
		end
		
		// draw percentage
		for j = damagestring:len(), 1, -1
			local lx
			local ly = y + DAMAGE_YOFFSET
			local color = SKINCOLOR_WHITE
			local letter = damagestring:sub(j, j)
			damagex = $ - v.nameTagWidth(letter)*FRACUNIT
			lx = damagex
			if damage > 99
				color = SKINCOLOR_CRIMSON
				lx = $ + 4*v.RandomFixed() - 2*FRACUNIT
				ly = $ + 4*v.RandomFixed() - 2*FRACUNIT
			end
			v.drawScaledNameTag(lx, ly, letter, V_SNAPTOBOTTOM|V_HUDTRANS, FRACUNIT,
				color, SKINCOLOR_BLACK)
		end
		x = $ + space
	end
end

local function MinecraftHUD(v, player, cam)
	// vars
	local index = player == secondarydisplayplayer and 2 or 1
	local drawcompass = valid(compasseyes[index])
	local mo = player.realmo
	local inventory = player.btfmc_inventory
	local x, y = GetHotbarCoords()
	local flags = V_SNAPTOBOTTOM|V_PERPLAYER
	
	// draw hotbar
	v.drawScaled(x, y, FRACUNIT, HOTBAR, flags|V_HUDTRANS)
	v.drawScaled(x, y, FRACUNIT, SLOTS, flags|V_HUDTRANSHALF)
	
	// if needed, locate an existing eye of ender
	if not drawcompass
	and mo and #mapeyes
		local closestdist = INT32_MAX
		
		for i, eye in ipairs(mapeyes)
			local dist = R_PointToDist2(mo.x, mo.y, eye.x, eye.y)
			if dist < closestdist
				compasseyes[index] = eye
				closestdist = dist
			end
		end
		
		drawcompass = true
	end
	
	// if an eye of ender is being tracked, draw the compass in the offhand slot
	if drawcompass
		local function GetCameraAngle()
			if cam.chase
				return cam.angle
			else
				return player.cmd.angleturn << 16
			end
		end
		
		local compasseye = compasseyes[index]
		local angle = -(R_PointToAngle2(mo.x, mo.y, compasseye.x, compasseye.y) - GetCameraAngle() + ANGLE_180)
		local angdiff = angle - compassangles[index]
		
		if abs(compassspeeds[index]) < COMPASS_DECEL
		and abs(angdiff) < COMPASS_DECEL
			compassangles[index] = angle
			compassspeeds[index] = 0
		elseif (compassspeeds[index] > 0) == (angdiff > 0)
			if angdiff > 0
				compassspeeds[index] = $ + COMPASS_ACCEL
			else
				compassspeeds[index] = $ - COMPASS_ACCEL
			end
		else
			if angdiff > 0
				compassspeeds[index] = $ + COMPASS_DECEL
			else
				compassspeeds[index] = $ - COMPASS_DECEL
			end
		end
		
		compassangles[index] = $ + compassspeeds[index]
		
		angle = COMPASS_PATCHES*AngleFixed(compassangles[index] + COMPASS_OFFSET)/(360*FRACUNIT)
		
		x = $ - (OFFHAND.width + 7)*FRACUNIT
		
		v.drawScaled(x, y, FRACUNIT, OFFHAND, flags|V_HUDTRANS)
		v.drawScaled(x, y, FRACUNIT, OFFSLOT, flags|V_HUDTRANSHALF)
		
		x = $ + 3*FRACUNIT
		y = $ + 3*FRACUNIT
		
		v.drawScaled(x, y, FRACUNIT, COMPASS[angle], flags|V_HUDTRANS)
	end
	
	// draw inventory
	if inventory
		local textx, texty = GetHotbarCoords()
		x, y = GetHotbarCoords()
		x = $ + (3 + 16 - 20)*FRACUNIT
		y = $ + (3 + 16)*FRACUNIT
		for i = 1, min(#inventory, INVENTORY_SLOTS)
			local entry = inventory[i]
			local patch
			local hscale = FRACUNIT
			local vscale = FRACUNIT
			
			x = $ + 20*FRACUNIT
			
			if entry.count == 0 then continue end
			
			patch = v.getSpritePatch(SPR_MCIT, entry.frame)
			
			if entry.squishtics
				hscale = $ - entry.squishtics*SQUISH_FACTOR/SQUISH_TICS
				vscale = $ + entry.squishtics*SQUISH_FACTOR/SQUISH_TICS
				entry.squishtics = $ - 1
			end
			
			v.drawStretched(x - 8*FRACUNIT, y + patch.height*vscale/4 - 8*FRACUNIT, hscale, vscale, patch, flags|V_HUDTRANS)
			if entry.throwdelay
				hscale = 16*FRACUNIT/WHITE.width
				vscale = 16*entry.throwdelay*FRACUNIT/(WHITE.height*THROW_DELAY)
				v.drawStretched(x - WHITE.width*hscale, y - WHITE.height*vscale, hscale, vscale, WHITE, flags|V_HUDTRANSHALF)
			end
			if entry.count > 1
				DrawMinecraftNumber(v, x, y, FRACUNIT, "MCFNT", entry.count, flags|V_HUDTRANS, SKINCOLOR_WHITE, true)
			end
		end
	end
end

hud.add(function(v, player, cam)
	if btfmc_hud == HUD_NONE
		return
	end
	
	// cache patches
	if not valid(HOTBAR)
		HOTBAR = v.cachePatch("MCHOTBAR")
		SLOTS = v.cachePatch("MCSLOTS")
		OFFHAND = v.cachePatch("MCOFHAND")
		OFFSLOT = v.cachePatch("MCOFSLOT")
		WHITE = v.cachePatch("~000")
		for i = 0, COMPASS_PATCHES
			COMPASS[i] = v.cachePatch("MCCMPS"..i)
		end
	end
	
	if btfmc_hud == HUD_MINECRAFT
		MinecraftHUD(v, player, cam)
	elseif btfmc_hud == HUD_SMASH
		SmashHUD(v, player, cam)
	end
end)